import httpx
import pytest
from pytest_httpx import HTTPXMock

from nextline_schedule.scheduler import RequestStatement

COMMANDS = '''
import time
# 2023-03-31 16:12:59
time.sleep(7.61)
time.sleep(5.45)
time.sleep(11.82)
time.sleep(35.12)
'''.lstrip()


async def test_success(httpx_mock: HTTPXMock) -> None:
    json = {
        'commands': COMMANDS,
        # 'message': 'Success',
        # 'status': 'ok',
    }
    httpx_mock.add_response(json=json)
    func = RequestStatement(
        api_url='https://scheduler-uobd.onrender.com/api/v1/schedule/',
        length_minutes=1,
        policy='dummy',
    )

    statement = await func()
    assert statement == COMMANDS


async def test_400(httpx_mock: HTTPXMock) -> None:
    httpx_mock.add_response(status_code=400)
    func = RequestStatement(
        api_url='https://scheduler-uobd.onrender.com/api/v1/schedule/',
        length_minutes=1,
        policy='dummy',
    )

    with pytest.raises(RuntimeError):
        await func()


async def test_timeout(httpx_mock: HTTPXMock) -> None:

    httpx_mock.add_exception(httpx.ReadTimeout('timeout'))

    func = RequestStatement(
        api_url='https://scheduler-uobd.onrender.com/api/v1/schedule/',
        length_minutes=1,
        policy='dummy',
    )

    with pytest.raises(httpx.TimeoutException):
        await func()


@pytest.mark.skip
async def test_with_dummy_server() -> None:
    func = RequestStatement(
        api_url='https://scheduler-uobd.onrender.com/api/v1/schedule/',
        length_minutes=1,
        policy='dummy',
    )

    statement = await func()
    print(statement)



from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.utils import safe_compare
from nextline_schedule.utils.strategies import st_datetimes, st_ranges


@given(st.data())
def test_st_datetimes(data: st.DataObject) -> None:
    min_, max_ = data.draw(st_ranges(st_=st_datetimes))
    dt_ = data.draw(st_datetimes(min_value=min_, max_value=max_))
    assert dt_.tzinfo is None
    assert dt_.fold == 0
    assert safe_compare(min_) <= dt_ <= safe_compare(max_)



import random
from datetime import datetime


class DummyRequestStatement:
    async def __call__(self) -> str:
        return generate_dummy_statement()


def generate_dummy_statement() -> str:
    s = '\n'.join(
        f'time.sleep({random.uniform(1, 3):.2})' for _ in range(random.randint(5, 10))
    )
    s = '\n'.join(
        (
            f'# {datetime.utcnow().isoformat()}',
            'import time',
            '',
            s,
        )
    )
    return s



import strawberry
from strawberry.types import Info

from nextline_schedule.queue import PubSubQueue, PushArg

from .types import ScheduleQueueItem


@strawberry.input
class ScheduleQueuePushInput:
    name: str
    script: str

    def to_push_arg(self) -> PushArg:
        return PushArg(
            name=self.name,
            script=self.script,
        )


async def mutate_push(info: Info, input: ScheduleQueuePushInput) -> ScheduleQueueItem:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, PubSubQueue)
    push_arg = input.to_push_arg()
    item = await queue.push(push_arg)
    return ScheduleQueueItem.from_(item)


async def mutate_remove(info: Info, id: int) -> bool:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, PubSubQueue)
    return await queue.remove(id)


@strawberry.type
class MutationScheduleQueue:
    push: ScheduleQueueItem = strawberry.field(resolver=mutate_push)
    remove: bool = strawberry.field(resolver=mutate_remove)



from pathlib import Path

pwd = Path(__file__).resolve().parent

sub = pwd / 'mutations'
MUTATE_AUTO_MODE_TURN_ON = (sub / 'AutoModeTurnOn.gql').read_text()
MUTATE_AUTO_MODE_TURN_OFF = (sub / 'AutoModeTurnOff.gql').read_text()
MUTATE_QUEUE_PUSH = (sub / 'QueuePush.gql').read_text()


sub = pwd / 'queries'
QUERY_AUTO_MODE = (sub / 'AutoMode.gql').read_text()
QUERY_SCHEDULER = (sub / 'Scheduler.gql').read_text()
QUERY_VERSION = (sub / 'Version.gql').read_text()
QUERY_SCHEDULE_QUEUE_ITEMS = (sub / 'QueueItems.gql').read_text()

sub = pwd / 'subscriptions'
SUBSCRIBE_AUTO_MODE_STATE = (sub / 'ScheduleAutoModeState.gql').read_text()



from collections.abc import AsyncIterator

from strawberry.types import Info

from nextline_schedule.auto import AutoMode


def subscribe_auto_mode_state(info: Info) -> AsyncIterator[str]:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    return auto_mode.subscribe_state()


def subscribe_auto_mode_mode(info: Info) -> AsyncIterator[str]:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    return auto_mode.subscribe_mode()



import strawberry

import nextline_schedule
from nextline_schedule.graphql import QUERY_VERSION
from nextline_schedule.schema import Query


async def test_version():
    schema = strawberry.Schema(query=Query)
    resp = await schema.execute(QUERY_VERSION)
    assert resp.data['schedule']['version'] == nextline_schedule.__version__



import asyncio
import time

from nextline import Nextline

from nextline_schedule.auto import AutoModeStateMachine, AutoMode

STATEMENT_QUEUE = '''
"""queue"""
import time
time.sleep(0.01)
'''


async def mock_queue() -> str:
    return STATEMENT_QUEUE


class MockError(Exception):
    pass


def f():
    time.sleep(0.001)


def g():
    time.sleep(0.001)
    raise MockError()


async def test_on_raised_while_pulling():
    async def pull():
        raise MockError()

    run_no = 1
    nextline = Nextline(statement=f, run_no_start_from=run_no, timeout_on_exit=60)
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=mock_queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await auto_mode.turn_on()
            async for state in auto_mode.subscribe_state():
                if state == 'off':
                    break

    expected = ['off', 'auto_waiting', 'auto_pulling', 'off']
    assert expected == await states


async def test_on_raised_while_running():
    async def pull():
        return g

    run_no = 1
    nextline = Nextline(statement=f, run_no_start_from=run_no, timeout_on_exit=60)
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=mock_queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await auto_mode.turn_on()
            async for state in auto_mode.subscribe_state():
                if state == 'off':
                    break

    expected = ['off', 'auto_waiting', 'auto_pulling', 'auto_running', 'off']
    assert expected == await states


async def subscribe_state(auto_mode: AutoModeStateMachine):
    return [state async for state in auto_mode.subscribe_state()]



import strawberry
from strawberry.types import Info

from nextline_schedule import __version__
from nextline_schedule.scheduler import RequestStatement

from .auto import QueryScheduleAutoMode
from .queue import QueryScheduleQueue


def query_scheduler_api_url(info: Info) -> str:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, RequestStatement)
    return scheduler._api_url


def query_scheduler_length_minutes(info: Info) -> int:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, RequestStatement)
    return scheduler._length_minutes


def query_scheduler_policy(info: Info) -> str:
    scheduler = info.context['schedule']['scheduler']
    assert isinstance(scheduler, RequestStatement)
    return scheduler._policy


@strawberry.type
class QueryScheduler:
    api_url: str = strawberry.field(resolver=query_scheduler_api_url)
    length_minutes: int = strawberry.field(resolver=query_scheduler_length_minutes)
    policy: str = strawberry.field(resolver=query_scheduler_policy)


@strawberry.type
class QuerySchedule:
    version: str = __version__

    @strawberry.field
    def auto_mode(self, info: Info) -> QueryScheduleAutoMode:
        return QueryScheduleAutoMode()

    @strawberry.field
    def scheduler(self, info: Info) -> QueryScheduler:
        return QueryScheduler()

    @strawberry.field
    def queue(self) -> QueryScheduleQueue:
        return QueryScheduleQueue()


@strawberry.type
class Query:
    @strawberry.field
    def schedule(self, info: Info) -> QuerySchedule:
        return QuerySchedule()



from typing import Optional, TypeVar

from hypothesis import given, settings
from hypothesis import strategies as st

from nextline_schedule.utils import safe_compare, safe_max
from nextline_schedule.utils.strategies import (
    StMinMaxValuesFactory,
    st_datetimes,
    st_graphql_ints,
    st_none_or,
    st_ranges,
)

T = TypeVar('T')


def st_min_max_start(
    st_: StMinMaxValuesFactory[T],
) -> st.SearchStrategy[tuple[Optional[T], Optional[T]]]:
    def st_min() -> st.SearchStrategy[Optional[T]]:
        return st_none_or(st_())

    def st_max(min_value: Optional[T]) -> st.SearchStrategy[Optional[T]]:
        return st_none_or(st_(min_value=min_value))

    return st_min().flatmap(lambda min_: st.tuples(st.just(min_), st_max(min_)))


def st_min_max_end(
    st_: StMinMaxValuesFactory[T],
    min_start: Optional[T] = None,
) -> st.SearchStrategy[tuple[Optional[T], Optional[T]]]:
    def st_min() -> st.SearchStrategy[Optional[T]]:
        return st_none_or(st_(min_value=min_start))

    def st_max(min_value: Optional[T]) -> st.SearchStrategy[Optional[T]]:
        min_value = safe_max((min_value, min_start))
        return st_none_or(st_(min_value=min_value))

    return st_min().flatmap(lambda min_: st.tuples(st.just(min_), st_max(min_)))


@given(st.data())
@settings(max_examples=1000)
def test_st_ranges(data: st.DataObject) -> None:
    st_ = data.draw(st.sampled_from([st_graphql_ints, st_datetimes]))

    min_start, max_start = data.draw(st_min_max_start(st_=st_))  # type: ignore
    min_end, max_end = data.draw(st_min_max_end(st_=st_, min_start=min_start))  # type: ignore

    assert safe_compare(min_start) <= safe_compare(max_start)
    assert safe_compare(min_start) <= safe_compare(min_end) <= safe_compare(max_end)

    allow_start_none = data.draw(st.booleans())
    allow_end_none = data.draw(st.booleans())
    allow_equal = data.draw(st.booleans())
    let_end_none_if_start_none = data.draw(st.booleans())

    start, end = data.draw(
        st_ranges(
            st_=st_,  # type: ignore
            min_start=min_start,
            max_start=max_start,
            min_end=min_end,
            max_end=max_end,
            allow_start_none=allow_start_none,
            allow_end_none=allow_end_none,
            allow_equal=allow_equal,
            let_end_none_if_start_none=let_end_none_if_start_none,
        )
    )

    if not allow_start_none:
        assert start is not None

    if start is None and let_end_none_if_start_none:
        assert end is None
    elif not allow_end_none:
        assert end is not None

    if allow_equal:
        assert safe_compare(start) <= safe_compare(end)
    else:
        assert safe_compare(start) < safe_compare(end)

    assert safe_compare(min_start) <= safe_compare(start) <= safe_compare(max_start)
    assert safe_compare(min_end) <= safe_compare(end) <= safe_compare(max_end)



import strawberry
from strawberry.types import Info

from nextline_schedule.queue import PubSubQueue

from .types import ScheduleQueueItem


async def resolve_items(info: Info) -> list[ScheduleQueueItem]:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, PubSubQueue)
    ret = [ScheduleQueueItem.from_(item) for item in queue.items]
    return ret


@strawberry.type
class QueryScheduleQueue:
    items: list[ScheduleQueueItem] = strawberry.field(resolver=resolve_items)



import asyncio
import time

from nextline import Nextline

from nextline_schedule.auto import AutoMode, AutoModeStateMachine

STATEMENT_QUEUE = '''
"""queue"""
import time
time.sleep(0.01)
'''


async def mock_queue() -> str:
    return STATEMENT_QUEUE


def f():
    time.sleep(0.001)


async def pull():
    return f


async def test_turn_off_while_waiting():
    run_no = 1
    nextline = Nextline(
        statement=f,
        run_no_start_from=run_no,
        trace_threads=True,
        trace_modules=True,
        timeout_on_exit=60,
    )
    auto_mode = AutoMode(nextline=nextline, scheduler=pull, queue=mock_queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await nextline.run()
            async for prompt in nextline.prompts():
                break
            await auto_mode.turn_on()
            await auto_mode.turn_off()
            await nextline.send_pdb_command(
                command='continue',
                prompt_no=prompt.prompt_no,
                trace_no=prompt.trace_no,
            )

    expected = ['off', 'auto_waiting', 'off']
    assert expected == await states


async def subscribe_state(auto_mode: AutoModeStateMachine):
    return [state async for state in auto_mode.subscribe_state()]



import asyncio
import sys
from collections.abc import AsyncIterable, AsyncIterator, Iterable
from typing import TypeVar

import pytest

T = TypeVar('T')

if not sys.version_info >= (3, 10):
    # https://github.com/python/typeshed/blob/main/stdlib/builtins.pyi
    def aiter(async_iterable: AsyncIterable[T]) -> AsyncIterator[T]:
        return async_iterable.__aiter__()

    async def anext(async_iterator: AsyncIterator[T]) -> T:
        return await async_iterator.__anext__()


async def aiterable(iterable: Iterable[T]) -> AsyncIterator[T]:
    '''Wrap iterable so can be used with "async for"'''
    for i in iterable:
        await asyncio.sleep(0)  # let other tasks run
        yield i


async def test_aiter__() -> None:
    agen = (i async for i in aiterable(range(5)))
    iter = agen.__aiter__()
    assert await iter.__anext__() == 0
    assert await iter.__anext__() == 1
    assert await iter.__anext__() == 2
    assert await iter.__anext__() == 3
    assert await iter.__anext__() == 4
    with pytest.raises(StopAsyncIteration):
        await iter.__anext__()


async def test_aiter() -> None:
    agen = (i async for i in aiterable(range(5)))
    iter = aiter(agen)
    assert await anext(iter) == 0
    assert await anext(iter) == 1
    assert await anext(iter) == 2
    assert await anext(iter) == 3
    assert await anext(iter) == 4
    with pytest.raises(StopAsyncIteration):
        await anext(iter)



import asyncio
from collections.abc import AsyncIterator

import pytest
from nextlinegraphql import create_app
from nextlinegraphql.plugins.ctrl.graphql import SUBSCRIBE_STATE
from nextlinegraphql.plugins.graphql.test import TestClient, gql_request, gql_subscribe
from pytest_httpx import HTTPXMock

from nextline_schedule.graphql import (
    MUTATE_AUTO_MODE_TURN_OFF,
    MUTATE_AUTO_MODE_TURN_ON,
    QUERY_AUTO_MODE,
    QUERY_SCHEDULER,
    SUBSCRIBE_AUTO_MODE_STATE,
)


@pytest.fixture
async def client() -> AsyncIterator[TestClient]:
    app = create_app()  # the plugin is loaded here
    async with TestClient(app) as y:
        await asyncio.sleep(0)
        yield y


async def test_plugin(client: TestClient):
    turned_on = asyncio.Event()
    task = asyncio.create_task(subscribe_auto_mode_state(client, turned_on))

    data = await gql_request(client, QUERY_SCHEDULER)
    expected = {
        'schedule': {
            'scheduler': {
                'apiUrl': 'https://example.com',
                'lengthMinutes': 60,
                'policy': 'test',
            }
        }
    }
    assert data == expected

    data = await gql_request(client, QUERY_AUTO_MODE)
    expected = {'schedule': {'autoMode': {'state': 'off'}}}
    assert data == expected

    data = await gql_request(client, MUTATE_AUTO_MODE_TURN_ON)

    n_runs = 0
    async for data in gql_subscribe(client, SUBSCRIBE_STATE):
        if data['ctrlState'] == 'running':
            turned_on.set()
            n_runs += 1
        if n_runs >= 3:
            data = await gql_request(client, MUTATE_AUTO_MODE_TURN_OFF)
            break

    async for data in gql_subscribe(client, SUBSCRIBE_STATE):
        if data['ctrlState'] == 'finished':
            break

    await task


async def subscribe_auto_mode_state(client: TestClient, turned_on: asyncio.Event):
    await turned_on.wait()
    ret = []
    async for data in gql_subscribe(client, SUBSCRIBE_AUTO_MODE_STATE):
        state = data['scheduleAutoModeState']
        ret.append(state)
        if state == 'off':
            break
    return ret


@pytest.fixture(autouse=True)
def configure_by_envvar(monkeypatch: pytest.MonkeyPatch):
    monkeypatch.setenv('NEXTLINE_SCHEDULE__API', 'https://example.com')
    monkeypatch.setenv('NEXTLINE_SCHEDULE__LENGTH_MINUTES', '60')
    monkeypatch.setenv('NEXTLINE_SCHEDULE__POLICY', 'test')


COMMANDS = '''
import time
time.sleep(0.001)
'''.lstrip()


@pytest.fixture(autouse=True)
def mock_httpx(httpx_mock: HTTPXMock):
    httpx_mock.add_response(json={'commands': COMMANDS})



import asyncio

from nextline import Nextline

from nextline_schedule.auto import AutoMode

STATEMENT_SCHEDULER = '''
"""run_no: {run_no}"""
import time
time.sleep(0.01)
'''

STATEMENT_QUEUE = '''
"""queue"""
import time
time.sleep(0.01)
'''


class MockScheduler:
    def __init__(self, nextline: Nextline):
        self._nextline = nextline

    async def __call__(self) -> str:
        return STATEMENT_SCHEDULER.format(run_no=self._nextline.run_no + 1)


async def mock_queue() -> str:
    return STATEMENT_QUEUE


async def test_one() -> None:
    run_no = 1
    statement = STATEMENT_SCHEDULER.format(run_no=run_no)
    nextline = Nextline(statement=statement, run_no_start_from=run_no)
    scheduler = MockScheduler(nextline=nextline)
    auto_mode = AutoMode(nextline=nextline, scheduler=scheduler, queue=mock_queue)

    states = asyncio.create_task(subscribe_state(auto_mode))

    async with auto_mode:
        async with nextline:
            await control(auto_mode, nextline)

    expected = [
        'off',
        'auto_waiting',
        'auto_pulling',
        'auto_running',
        'auto_pulling',
        'auto_running',
        'off',
    ]
    assert expected == await states


async def control(auto_mode: AutoMode, nextline: Nextline):
    assert auto_mode.state == 'off'  # type: ignore
    await auto_mode.turn_on()  # type: ignore
    async for run_info in nextline.subscribe_run_info():
        if run_info.run_no == 3 and run_info.state == 'running':
            break
    await auto_mode.turn_off()  # type: ignore


async def subscribe_state(auto_mode: AutoMode):
    return [state async for state in auto_mode.subscribe_state()]



__all__ = [
    'st_datetimes',
    'StMinMaxValuesFactory',
    'st_graphql_ints',
    'st_none_or',
    'st_ranges',
    'st_python_scripts',
]


from .datetime import st_datetimes
from .misc import StMinMaxValuesFactory, st_graphql_ints, st_none_or, st_ranges
from .script import st_python_scripts



from datetime import datetime, timezone


def utc_timestamp(naive: datetime) -> int:
    '''Return the timestamp of the naive datetime object as UTC.

    https://docs.python.org/3/library/datetime.html#datetime.datetime.timestamp

    >>> utc_timestamp(datetime(2022, 10, 30, 17, 13, 20))
    1667150000

    '''
    assert not is_timezone_aware(naive)
    aware = naive.replace(tzinfo=timezone.utc)
    return int(aware.timestamp())


def is_timezone_aware(dt: datetime) -> bool:
    '''Return True if the datetime object has a timezone.

    https://docs.python.org/3/library/datetime.html#determining-if-an-object-is-aware-or-naive

    >>> is_timezone_aware(datetime(2022, 10, 30, 17, 13, 20))
    False

    >>> is_timezone_aware(datetime(2022, 10, 30, 17, 13, 20, tzinfo=timezone.utc))
    True

    '''
    return (dt.tzinfo and dt.tzinfo.utcoffset(dt)) is not None



from typing import Optional

import strawberry
from nextline import Nextline
from strawberry.types import Info

from nextline_schedule.scheduler import RequestStatement

from .auto import MutationScheduleAutoMode
from .queue import MutationScheduleQueue


@strawberry.input
class MutationSchedulerInput:
    api_url: Optional[str] = None
    length_minutes: Optional[int] = None
    policy: Optional[str] = None


@strawberry.type
class MutationScheduler:
    @strawberry.mutation
    def update(self, info: Info, input: MutationSchedulerInput) -> bool:
        scheduler = info.context['schedule']['scheduler']
        assert isinstance(scheduler, RequestStatement)
        if input.api_url is not None:
            scheduler._api_url = input.api_url
        if input.length_minutes is not None:
            scheduler._length_minutes = input.length_minutes
        if input.policy is not None:
            scheduler._policy = input.policy
        return True


async def mutate_load_script(info: Info) -> bool:
    nextline = info.context["nextline"]
    assert isinstance(nextline, Nextline)
    scheduler = info.context['schedule']['scheduler']
    assert callable(scheduler)
    statement = await scheduler()
    assert isinstance(statement, str)
    await nextline.reset(statement=statement)
    return True


@strawberry.type
class MutationSchedule:
    @strawberry.field
    def auto_mode(self, info: Info) -> MutationScheduleAutoMode:
        return MutationScheduleAutoMode()

    @strawberry.field
    def scheduler(self, info: Info) -> MutationScheduler:
        return MutationScheduler()

    @strawberry.mutation
    async def load_script(self, info: Info) -> bool:
        return await mutate_load_script(info)

    @strawberry.field
    def queue(self) -> MutationScheduleQueue:
        return MutationScheduleQueue()


@strawberry.type
class Mutation:
    @strawberry.field
    def schedule(self, info: Info) -> MutationSchedule:
        return MutationSchedule()



__all__ = [
    'MutationScheduleQueue',
    'QueryScheduleQueue',
    'subscribe_schedule_queue_items',
    'ScheduleQueueItem',
]

from .mutation import MutationScheduleQueue
from .query import QueryScheduleQueue
from .subscription import subscribe_schedule_queue_items
from .types import ScheduleQueueItem



from collections.abc import AsyncIterator

from strawberry.types import Info

from nextline_schedule.queue import PubSubQueue

from .types import ScheduleQueueItem


async def subscribe_schedule_queue_items(
    info: Info,
) -> AsyncIterator[list[ScheduleQueueItem]]:
    queue = info.context['schedule']['queue']
    assert isinstance(queue, PubSubQueue)
    async for items in queue.subscribe():
        yield [ScheduleQueueItem.from_(item) for item in items]



import dataclasses
from collections.abc import AsyncIterator, Awaitable, Callable
from typing import Literal

from nextline import Nextline
from nextline.utils import pubsub

from nextline_schedule.types import Statement

from .callback import Callback, build_state_machine
from .types import PullFunc

FromName = Literal['scheduler', 'queue']
ModeName = Literal['off', 'scheduler', 'queue']


class AutoMode:
    def __init__(self, nextline: Nextline, scheduler: PullFunc, queue: PullFunc):
        self._pull_from = PullFrom(
            scheduler=scheduler, queue=queue, on_from_changed=self.on_from_changed
        )
        callback = Callback(
            nextline=nextline,
            pull_func=self._pull_from,
            on_state_changed=self.on_state_changed,
        )
        self._machine = build_state_machine(nextline=nextline, callback=callback)
        self._mode: ModeName = 'off'
        self._pubsub_mode = pubsub.PubSubItem[ModeName]()

    @property
    def state(self) -> str:
        return self._machine.state

    def subscribe_state(self) -> AsyncIterator[str]:
        return self._machine.subscribe_state()

    @property
    def mode(self) -> ModeName:
        return self._mode

    def subscribe_mode(self) -> AsyncIterator[ModeName]:
        return self._pubsub_mode.subscribe()

    async def change_mode(self, mode: ModeName) -> None:
        if mode == 'off':
            await self._machine.turn_off()  # type: ignore
            return
        await self._pull_from.change_from(mode)
        await self._machine.turn_on()  # type: ignore

    async def turn_on(self) -> None:
        await self._machine.turn_on()  # type: ignore

    async def turn_off(self) -> None:
        await self._machine.turn_off()  # type: ignore

    async def __aenter__(self) -> 'AutoMode':
        await self._pubsub_mode.publish(self._mode)
        await self._machine.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        await self._machine.__aexit__(exc_type, exc_value, traceback)
        await self._pubsub_mode.close()

    async def on_state_changed(self, state: str) -> None:
        if state == 'off':
            if self._mode == 'off':
                return
            self._mode = 'off'
            await self._pubsub_mode.publish(self._mode)
            return

        if state.startswith('auto_'):
            from_ = self._pull_from.current
            if self._mode == from_:
                return
            self._mode = from_
            await self._pubsub_mode.publish(self._mode)

    async def on_from_changed(self, name: FromName) -> None:
        if self.state == 'off':
            return

        if self._mode == name:
            return
        self._mode = name
        await self._pubsub_mode.publish(self._mode)


@dataclasses.dataclass
class PullFromItem:
    name: FromName
    pull: PullFunc


class PullFrom:
    def __init__(
        self,
        scheduler: PullFunc,
        queue: PullFunc,
        on_from_changed: Callable[[FromName], Awaitable[None]],
    ):
        self._scheduler = PullFromItem(name='scheduler', pull=scheduler)
        self._queue = PullFromItem(name='queue', pull=queue)
        self._on_from_changed = on_from_changed
        self._map = {
            self._scheduler.name: self._scheduler,
            self._queue.name: self._queue,
        }
        self._current = self._scheduler

    async def __call__(self) -> Statement:
        return await self._current.pull()

    async def change_from(self, name: FromName) -> None:
        old = self._current.name
        self._current = self._map[name]
        if old != self._current.name:
            await self._on_from_changed(self._current.name)

    @property
    def current(self) -> FromName:
        return self._current.name



from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.utils import safe_compare
from nextline_schedule.utils.strategies import st_ranges


@given(st.data())
def test_safe_compare(data: st.DataObject) -> None:
    allow_equal = data.draw(st.booleans())
    none_or_small, none_or_large = data.draw(
        st_ranges(st_=st.integers, allow_equal=allow_equal)
    )
    if allow_equal:
        assert safe_compare(none_or_small) <= safe_compare(none_or_large)
        assert safe_compare(none_or_large) >= safe_compare(none_or_small)
    else:
        assert safe_compare(none_or_small) < safe_compare(none_or_large)
        assert safe_compare(none_or_large) > safe_compare(none_or_small)



from datetime import datetime, timedelta
from logging import getLogger

import httpx


class RequestStatement:
    def __init__(self, api_url: str, length_minutes: int, policy: str, timeout=5.0):
        self._api_url = api_url
        self._length_minutes = length_minutes
        self._policy = policy
        self._timeout = timeout
        self._logger = getLogger(__name__)

    async def __call__(self) -> str:
        # https://github.com/simonsobs/so-scheduler/blob/master/readme.md#schedule-api
        start_time = datetime.utcnow()
        duration = timedelta(minutes=self._length_minutes)
        end_time = start_time + duration
        start_time_str = start_time.strftime('%Y-%m-%d %H:%M')
        end_time_str = end_time.strftime('%Y-%m-%d %H:%M')
        data = {"t0": start_time_str, "t1": end_time_str, "policy": self._policy}
        self._logger.info(f'Pulling a script: {data!r}')
        try:
            async with httpx.AsyncClient() as client:
                response = await client.post(
                    self._api_url, json=data, timeout=self._timeout
                )
        except Exception:
            self._logger.exception('')
            raise
        if not response.status_code == 200:
            msg = f'Failed: {response.status_code!r}, {response.text!r}'
            self._logger.error(msg)
            raise RuntimeError(msg)
        self._logger.info(f'Response: {response.json()!r}')
        return response.json()['commands']



import strawberry
from strawberry.types import Info

from nextline_schedule.auto import AutoMode


async def mutate_turn_on(info: Info) -> bool:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    await auto_mode.turn_on()
    return True


async def mutate_turn_off(info: Info) -> bool:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    await auto_mode.turn_off()
    return True


async def mutate_change_mode(info: Info, mode: str) -> bool:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    if mode not in ('off', 'scheduler', 'queue'):
        return False
    await auto_mode.change_mode(mode)  # type: ignore
    return True


@strawberry.type
class MutationScheduleAutoMode:
    turn_on: bool = strawberry.field(resolver=mutate_turn_on)
    turn_off: bool = strawberry.field(resolver=mutate_turn_off)
    change_mode: bool = strawberry.field(resolver=mutate_change_mode)



from collections.abc import Mapping, MutableMapping
from logging import getLogger
from pathlib import Path
from typing import Optional

from apluggy import asynccontextmanager
from dynaconf import Dynaconf, Validator
from nextlinegraphql.hook import spec

from .__about__ import __version__
from .auto import AutoMode
from .dummy import DummyRequestStatement
from .queue import PubSubQueue
from .scheduler import RequestStatement
from .schema import Mutation, Query, Subscription

HERE = Path(__file__).resolve().parent
DEFAULT_CONFIG_PATH = HERE / 'default.toml'

assert DEFAULT_CONFIG_PATH.is_file()

PRELOAD = (str(DEFAULT_CONFIG_PATH),)
SETTINGS = ()
VALIDATORS = (
    Validator("SCHEDULE.API", must_exist=True, is_type_of=str),
    Validator("SCHEDULE.LENGTH_MINUTES", must_exist=True, is_type_of=int),
    Validator("SCHEDULE.POLICY", must_exist=True, is_type_of=str),
)


class Plugin:
    @spec.hookimpl
    def dynaconf_preload(self) -> Optional[tuple[str, ...]]:
        return PRELOAD

    @spec.hookimpl
    def dynaconf_settings_files(self) -> Optional[tuple[str, ...]]:
        return SETTINGS

    @spec.hookimpl
    def dynaconf_validators(self) -> Optional[tuple[Validator, ...]]:
        return VALIDATORS

    @spec.hookimpl
    def configure(self, settings: Dynaconf):
        logger = getLogger(__name__)
        logger.info(f'{__package__} version: {__version__}')
        api_rul = settings.schedule.api
        length_minutes = settings.schedule.length_minutes
        policy = settings.schedule.policy

        self._dummy = DummyRequestStatement()
        self._scheduler = RequestStatement(
            api_url=api_rul, length_minutes=length_minutes, policy=policy
        )
        # self._scheduler = self._dummy

    @spec.hookimpl
    def schema(self):
        return (Query, Mutation, Subscription)

    @spec.hookimpl
    @asynccontextmanager
    async def lifespan(self, context: Mapping):
        nextline = context['nextline']
        self._queue = PubSubQueue()
        self._auto_mode = AutoMode(
            nextline=nextline, scheduler=self._scheduler, queue=self._queue
        )
        async with self._queue, self._auto_mode as y:
            yield y

    @spec.hookimpl
    def update_strawberry_context(self, context: MutableMapping) -> None:
        schedule = {
            'auto_mode': self._auto_mode,
            'scheduler': self._scheduler,
            'queue': self._queue,
        }
        context['schedule'] = schedule



import strawberry
from hypothesis import given

from nextline_schedule.graphql import QUERY_SCHEDULE_QUEUE_ITEMS
from nextline_schedule.queue import PubSubQueue
from nextline_schedule.queue.strategies import st_pubsub_queue
from nextline_schedule.schema import Query


@given(queue=st_pubsub_queue())
async def test_query(queue: PubSubQueue):
    async with queue:
        context_schedule = {'queue': queue}
        context = {'schedule': context_schedule}
        schema = strawberry.Schema(query=Query)
        resp = await schema.execute(QUERY_SCHEDULE_QUEUE_ITEMS, context_value=context)
        assert resp.data
        assert len(resp.data['schedule']['queue']['items']) == len(queue.items)



__all__ = [
    'safe_compare',
    'safe_max',
    'safe_min',
    'is_timezone_aware',
    'utc_timestamp',
]

from .safe import safe_compare, safe_max, safe_min
from .utc import is_timezone_aware, utc_timestamp



__all__ = [
    'MutationScheduleAutoMode',
    'QueryScheduleAutoMode',
    'subscribe_auto_mode_mode',
    'subscribe_auto_mode_state',
]

from .mutation import MutationScheduleAutoMode
from .query import QueryScheduleAutoMode
from .subscription import subscribe_auto_mode_mode, subscribe_auto_mode_state



from copy import deepcopy
from typing import Any
from unittest.mock import AsyncMock

import pytest
from hypothesis import given
from hypothesis import strategies as st
from transitions import Machine
from transitions.extensions.markup import HierarchicalMarkupMachine

from nextline_schedule.auto.state_machine.factory import build_state_machine


def test_model_default():
    machine = build_state_machine()
    assert not machine.models


def test_model_self_literal():
    machine = build_state_machine(model=Machine.self_literal)
    assert machine.models[0] is machine
    assert len(machine.models) == 1


def test_restore_from_markup():
    machine = build_state_machine(markup=True)
    assert isinstance(machine.markup, dict)
    markup = deepcopy(machine.markup)
    del markup['models']
    rebuild = HierarchicalMarkupMachine(model=None, **markup)
    assert rebuild.markup == machine.markup


@pytest.mark.skip
def test_graph():
    machine = build_state_machine(model=Machine.self_literal, graph=True)
    machine.get_graph().draw('states.png', prog='dot')


@st.composite
def st_paths(draw: st.DrawFn):
    max_n_paths = draw(st.integers(min_value=1, max_value=30))

    state_map = {
        'created': {
            'start': {'dest': 'off'},
        },
        'off': {
            'turn_on': {'dest': 'auto_waiting'},
        },
        'auto_waiting': {
            'turn_off': {'dest': 'off', 'before': 'cancel_task'},
            'on_initialized': {'dest': 'auto_pulling'},
            'on_finished': {'dest': 'auto_pulling'},
            'on_raised': {'dest': 'off'},
        },
        'auto_pulling': {
            'run': {'dest': 'auto_running'},
            'turn_off': {'dest': 'off', 'before': 'cancel_task'},
            'on_raised': {'dest': 'off'},
        },
        'auto_running': {
            'on_finished': {'dest': 'auto_pulling'},
            'turn_off': {'dest': 'off', 'before': 'cancel_task'},
            'on_raised': {'dest': 'off'},
        },
    }
    final_states = {'off'}
    backwards = {'on_initialized', 'on_finished'}

    all_triggers = list({trigger for v in state_map.values() for trigger in v.keys()})

    state_map_reduced = {
        state: {trigger: v2 for trigger, v2 in v.items() if trigger not in backwards}
        for state, v in state_map.items()
    }

    paths: list[tuple[str, dict[str, Any]]] = []

    state = 'created'
    while len(paths) < max_n_paths:
        trigger_map = state_map[state]
        trigger = draw(st.sampled_from(all_triggers))
        if trigger in trigger_map:
            paths.append((trigger, trigger_map[trigger]))
            state = trigger_map[trigger]['dest']
        else:
            paths.append((trigger, {'invalid': True}))

    while state not in final_states:
        trigger_map = state_map_reduced[state]
        triggers = list(trigger_map.keys())
        trigger = draw(st.sampled_from(triggers))
        paths.append((trigger, trigger_map[trigger]))
        state = trigger_map[trigger]['dest']

    return paths


@given(paths=st_paths())
async def test_transitions_hypothesis(paths: list[tuple[str, dict[str, Any]]]):
    machine = build_state_machine(model=Machine.self_literal)
    assert machine.is_created()

    for method, map in paths:
        if map.get('invalid'):
            await getattr(machine, method)()
            continue

        if before := map.get('before'):
            setattr(machine, before, AsyncMock())

        await getattr(machine, method)()
        dest = map['dest']
        assert getattr(machine, f'is_{dest}')()

        if before:
            assert getattr(machine, before).call_count == 1
            assert getattr(machine, before).await_count == 1



import asyncio

from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.queue import QueueItem
from nextline_schedule.queue.strategies import st_pubsub_queue, st_push_arg


@given(st.data())
async def test_pubsub_queue(data: st.DataObject):
    queue = data.draw(st_pubsub_queue())

    async def subscribe() -> list[list[QueueItem]]:
        return [i async for i in queue.subscribe()]

    task = asyncio.create_task(subscribe())
    await asyncio.sleep(0)

    methods = ['push', 'pop', 'remove', 'break']
    cmds = data.draw(st.lists(st.sampled_from(methods), min_size=0, max_size=10))

    async with queue:
        expected = [list(queue.items)]
        for cmd in cmds:
            await asyncio.sleep(0)
            match cmd:
                case 'push':
                    arg = data.draw(st_push_arg())
                    await queue.push(arg)
                case 'pop':
                    item = await queue.pop()
                    if item is None:
                        continue
                case 'remove':
                    ids = [i.id for i in queue.items]
                    ids.append(st.integers(min_value=0))
                    id = data.draw(st.sampled_from(ids))
                    success = await queue.remove(id)
                    if not success:
                        continue
                case 'break':
                    break
            expected.append(list(queue.items))

    await asyncio.sleep(0)

    actual = await task
    assert actual == expected



from typing import Optional

from hypothesis import strategies as st

from nextline_schedule.utils.strategies import st_datetimes, st_python_scripts

from .pubsub_queue import PubSubQueue
from .queue import PushArg, Queue, QueueItem


def st_queue_item(id: int) -> st.SearchStrategy[QueueItem]:
    return st.builds(
        QueueItem,
        id=st.just(id),
        name=st.text(),
        created_at=st_datetimes(),
        script=st_python_scripts(),
    )


def st_push_arg() -> st.SearchStrategy[PushArg]:
    return st.builds(
        PushArg,
        name=st.text(),
        script=st_python_scripts(),
    )


@st.composite
def st_queue_item_list(
    draw: st.DrawFn, min_size: int = 0, max_size: Optional[int] = None
) -> list[QueueItem]:
    size = draw(st.integers(min_value=min_size, max_value=max_size))
    return [draw(st_queue_item(id=i)) for i in range(size)]


def st_queue(
    min_items_size: int = 0, max_items_size: int = 5
) -> st.SearchStrategy[Queue]:
    return st.builds(
        Queue,
        items=st_queue_item_list(min_size=min_items_size, max_size=max_items_size),
    )


def st_pubsub_queue(
    min_items_size: int = 0, max_items_size: int = 5
) -> st.SearchStrategy[PubSubQueue]:
    return st.builds(
        PubSubQueue,
        items=st_queue_item_list(min_size=min_items_size, max_size=max_items_size),
    )



import strawberry
from strawberry.types import Info

from nextline_schedule.auto import AutoMode


def query_auto_mode_state(info: Info) -> str:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    return auto_mode.state


def query_auto_mode_mode(info: Info) -> str:
    auto_mode = info.context['schedule']['auto_mode']
    assert isinstance(auto_mode, AutoMode)
    return auto_mode.mode


@strawberry.type
class QueryScheduleAutoMode:
    state: str = strawberry.field(resolver=query_auto_mode_state)
    mode: str = strawberry.field(resolver=query_auto_mode_mode)



import strawberry
from hypothesis import given
from hypothesis import strategies as st

from nextline_schedule.graphql import MUTATE_QUEUE_PUSH
from nextline_schedule.queue import PubSubQueue
from nextline_schedule.queue.strategies import st_pubsub_queue, st_push_arg
from nextline_schedule.schema import Mutation, Query


@st.composite
def st_input(draw: st.DrawFn) -> dict:
    push_arg = draw(st_push_arg())
    return {'name': push_arg.name, 'script': push_arg.script}


@given(queue=st_pubsub_queue(), input=st_input())
async def test_query(queue: PubSubQueue, input: dict):
    async with queue:
        initial_len = len(queue.items)
        context_schedule = {'queue': queue}
        context = {'schedule': context_schedule}
        variable = {'input': input}
        schema = strawberry.Schema(query=Query, mutation=Mutation)
        resp = await schema.execute(
            MUTATE_QUEUE_PUSH, context_value=context, variable_values=variable
        )
        assert resp.data
        ret = resp.data['schedule']['queue']['push']
        assert ret['script'] == input['script']
        assert len(queue.items) == initial_len + 1



from typing import Type

from transitions import Machine
from transitions.extensions import MachineFactory
from transitions.extensions.markup import HierarchicalMarkupMachine


def build_state_machine(model=None, graph=False, asyncio=True, markup=False) -> Machine:
    '''Finite state machine for the auto-pilot mode states.

    State Diagram:

                         .-------------.
                         |   Created   |
                         '-------------'
                                | start()
                                |
                                v
                         .-------------.
         .-------------->|     Off     |<--------------.
         |               '-------------'               |
         |                      | turn_on()            |
       turn_off()               |                 on_raised()
         |                      |                      |
         |                      |                      |
         |   .------------------+------------------.   |
         |   |   Auto           |                  |   |
         |   |                  v                  |   |
         |   |           .-------------.           |   |
         |   |           |   Waiting   |           |   |
         |   |           '-------------'           |   |
         |   |                  | on_initialized() |   |
         |   |                  | on_finished()    |   |
         |   |                  v                  |   |
         |   |           .-------------.           |   |
         |   |           |   Pulling   |           |   |
         '---|           '-------------'           |---'
             |         run() |     ^               |
             |               |     |               |
             |               v     | on_finished() |
             |           .-------------.           |
             |           |   Running   |           |
             |           '-------------'           |
             |                                     |
             '-------------------------------------'

    >>> class Model:
    ...     def on_enter_auto_waiting(self):
    ...         print('enter the waiting state')
    ...         self.on_finished()
    ...
    ...     def on_exit_auto_waiting(self):
    ...        print('exit the waiting state')
    ...
    ...     def on_enter_auto_pulling(self):
    ...        print('enter the pulling state')

    >>> model = Model()
    >>> machine = build_state_machine(model=model, asyncio=False)
    >>> model.state
    'created'

    >>> _ = model.start()
    >>> model.state
    'off'

    >>> _ = model.turn_on()
    enter the waiting state
    exit the waiting state
    enter the pulling state

    >>> model.state
    'auto_pulling'


    '''

    MachineClass: Type[Machine]
    if markup:
        MachineClass = HierarchicalMarkupMachine
    else:
        MachineClass = MachineFactory.get_predefined(
            graph=graph, nested=True, asyncio=asyncio
        )

    # Build a state machine with nested states from dict configs with "remap"
    # in the way described in the docs:
    # https://github.com/pytransitions/transitions/tree/0.9.0#reuse-of-previously-created-hsms

    auto_state_conf = {
        'name': 'auto',
        'states': ['waiting', 'pulling', 'running'],
        'transitions': [
            ['on_initialized', 'waiting', 'pulling'],
            ['on_finished', 'waiting', 'pulling'],
            ['run', 'pulling', 'running'],
            ['on_finished', 'running', 'pulling'],
        ],
        'initial': 'waiting',
        'queued': True,
        'ignore_invalid_triggers': True,
    }

    # Ideally, we would be able to pass the auto_state_conf dict directly to
    # the MachineClass constructor, but this doesn't work with "remap."
    # Instead, we have to create a new instance of the MachineClass and pass it
    # to the parent state machine.
    auto_state = MachineClass(model=None, **auto_state_conf)  # type: ignore

    state_conf = {
        'name': 'global',
        'states': [
            'created',
            'off',
            {'name': 'auto', 'children': auto_state},
        ],
        'transitions': [
            ['start', 'created', 'off'],
            ['turn_on', 'off', 'auto'],
            ['on_raised', 'auto', 'off'],
            {
                'trigger': 'turn_off',
                'source': 'auto',
                'dest': 'off',
                'before': 'cancel_task',
            },
        ],
        'initial': 'created',
        'queued': True,
        'ignore_invalid_triggers': True,
    }

    machine = MachineClass(model=model, **state_conf)  # type: ignore
    return machine



from collections.abc import AsyncIterator, Iterable
from logging import getLogger
from typing import Optional

from nextline.utils import pubsub

from .queue import PushArg, Queue, QueueItem


class QueueEmpty(Exception):
    pass


class PubSubQueue:
    def __init__(self, items: Optional[Iterable[QueueItem]] = None):
        self._pubsub = pubsub.PubSubItem[list[QueueItem]]()
        self._queue = Queue(items)
        self._logger = getLogger(__name__)

    async def __call__(self) -> str:
        item = await self.pop()
        if item is None:
            self._logger.info('Queue is empty')
            raise QueueEmpty
        return item.script

    @property
    def items(self):
        return self._queue.items

    def subscribe(self) -> AsyncIterator[list[QueueItem]]:
        return self._pubsub.subscribe()

    def __len__(self) -> int:
        return len(self._queue)

    async def push(self, arg: PushArg) -> QueueItem:
        item = self._queue.push(arg)
        await self._pubsub.publish(list(self._queue.items))
        return item

    async def pop(self) -> QueueItem | None:
        item = self._queue.pop()
        if item is None:
            return None
        await self._pubsub.publish(list(self._queue.items))
        return item

    async def remove(self, id: int) -> bool:
        if not self._queue.remove(id):
            return False
        await self._pubsub.publish(list(self._queue.items))
        return True

    async def aclose(self) -> None:
        await self._pubsub.close()

    async def __aenter__(self) -> 'PubSubQueue':
        await self._pubsub.publish(list(self._queue.items))
        return self

    async def __aexit__(self, *_, **__) -> None:
        await self.aclose()



import datetime

import strawberry

from nextline_schedule.queue import QueueItem


@strawberry.type
class ScheduleQueueItem:
    id: int
    name: str
    created_at: datetime.datetime
    script: str

    @classmethod
    def from_(cls, src: QueueItem) -> 'ScheduleQueueItem':
        return cls(
            id=src.id,
            name=src.name,
            created_at=src.created_at,
            script=src.script,
        )



import asyncio
from collections.abc import AsyncIterator, Callable
from logging import getLogger
from typing import Any, Protocol

from nextline.utils import pubsub

from .factory import build_state_machine


class CallbackType(Protocol):
    async def on_state_changed(self, state: str) -> None:
        ...

    async def wait(self) -> None:
        ...

    async def pull(self) -> None:
        ...

    async def run(self, started: asyncio.Event) -> None:
        ...


class AutoModeStateMachine:
    '''The finite state machine for auto mode.'''

    def __init__(self, callback: CallbackType):
        self._callback = callback
        self._tasks = set[asyncio.Task]()
        self._pubsub_state = pubsub.PubSubItem[str]()
        self._logger = getLogger(__name__)

        machine = build_state_machine(model=self)
        machine.after_state_change = [self.after_state_change.__name__]

        self.state: str  # attached by machine
        self.start: Callable[..., Any]  # attached by machine

        assert isinstance(self.state, str)
        assert callable(self.start)

    def subscribe_state(self) -> AsyncIterator[str]:
        return self._pubsub_state.subscribe()

    async def on_enter_auto_waiting(self) -> None:
        task = asyncio.create_task(self._callback.wait())
        self._task = task
        self._tasks.add(task)

    async def on_enter_auto_pulling(self) -> None:
        task = asyncio.create_task(self._callback.pull())
        self._task = task
        self._tasks.add(task)

    async def on_enter_auto_running(self) -> None:
        started = asyncio.Event()
        task = asyncio.create_task(self._callback.run(started=started))
        await started.wait()
        self._task = task
        self._tasks.add(task)

    async def after_state_change(self) -> None:
        await self._collect_tasks()
        await self._callback.on_state_changed(self.state)
        await self._pubsub_state.publish(self.state)

    async def cancel_task(self) -> None:
        self._task.cancel()

    async def _collect_tasks(self) -> None:
        if not self._tasks:
            return
        done, _ = await asyncio.wait(
            self._tasks,
            timeout=0,
            return_when=asyncio.FIRST_COMPLETED,
        )
        for task in done:
            if (exc := task.exception()) is not None:
                self._logger.warning('An exception raised in a task', exc_info=exc)
        self._tasks -= done

    async def close(self) -> None:
        await self._collect_tasks()
        await self._pubsub_state.close()

    async def __aenter__(self) -> 'AutoModeStateMachine':
        await self.start()
        return self

    async def __aexit__(self, exc_type, exc_value, traceback) -> None:
        del exc_type, exc_value, traceback
        await self.close()



from hypothesis import given

from nextline_schedule.queue import PushArg, Queue
from nextline_schedule.queue.strategies import st_push_arg, st_queue


@given(queue=st_queue())
def test_queue(queue: Queue):
    # __len__
    assert len(queue) == len(queue.items)

    # Unique ids
    assert len(set(i.id for i in queue.items)) == len(queue)


@given(queue=st_queue(), push_arg=st_push_arg())
def test_push(queue: Queue, push_arg: PushArg):
    initial_len = len(queue)
    item = queue.push(push_arg)
    assert item.script == push_arg.script
    assert len(queue) == initial_len + 1
    assert len(set(i.id for i in queue.items)) == len(queue)


@given(queue=st_queue())
def test_pop(queue: Queue):
    initial_len = len(queue)
    if initial_len == 0:
        assert queue.pop() is None
        return
    item = queue.pop()
    assert len(queue) == initial_len - 1
    assert item not in queue.items


@given(queue=st_queue())
def test_remove(queue: Queue):
    initial_len = len(queue)
    if initial_len == 0:
        assert not queue.remove(0)
        return
    item = queue.items[0]
    assert queue.remove(item.id)
    assert len(queue) == initial_len - 1
    assert item not in queue.items
    assert not queue.remove(item.id)
    assert len(queue) == initial_len - 1
    assert item not in queue.items



import asyncio
from collections.abc import Awaitable, Callable
from logging import getLogger

from nextline import Nextline
from nextline.plugin.spec import Context, hookimpl

from .state_machine.machine import AutoModeStateMachine
from .types import PullFunc


def build_state_machine(
    nextline: Nextline, callback: 'Callback'
) -> AutoModeStateMachine:
    machine = AutoModeStateMachine(callback=callback)
    callback.machine = machine
    plugin = ScheduleAutoMode(machine=machine)
    nextline.register(plugin)
    return machine


class ScheduleAutoMode:
    '''A plugin for Nextline.

    The name of this class appears as the plugin name in the log.
    '''

    def __init__(self, machine: AutoModeStateMachine):
        self._machine = machine

    @hookimpl
    async def on_initialize_run(self) -> None:
        await self._machine.on_initialized()  # type: ignore

    @hookimpl
    async def on_finished(self, context: Context) -> None:
        nextline = context.nextline
        if nextline.format_exception():
            await self._machine.on_raised()  # type: ignore
            return
        await self._machine.on_finished()  # type: ignore


class Callback:
    def __init__(
        self,
        nextline: Nextline,
        pull_func: PullFunc,
        on_state_changed: Callable[[str], Awaitable[None]],
    ):
        self._nextline = nextline
        self._pull = pull_func
        self._on_state_changed = on_state_changed
        self._logger = getLogger(__name__)
        self.machine: AutoModeStateMachine  # to be set

    async def on_state_changed(self, state: str) -> None:
        await self._on_state_changed(state)

    async def wait(self) -> None:
        match self._nextline.state:
            case 'initialized':
                await self.machine.on_initialized()  # type: ignore
            case 'finished':
                await self.machine.on_finished()  # type: ignore

    async def pull(self) -> None:
        try:
            try:
                statement = await self._pull()
            except Exception:
                self._logger.exception('')
                await self.machine.on_raised()  # type: ignore
                return
            await self._nextline.reset(statement=statement)
            await self.machine.run()  # type: ignore
        except asyncio.CancelledError:
            self._logger.info(f'{self.__class__.__name__}.pull() cancelled')

    async def run(self, started: asyncio.Event) -> None:
        try:
            await self._nextline.run_continue_and_wait(started)
        except asyncio.CancelledError:
            self._logger.info(f'{self.__class__.__name__}.run() cancelled')



from collections.abc import AsyncIterator

import strawberry

from .auto import subscribe_auto_mode_mode, subscribe_auto_mode_state
from .queue import ScheduleQueueItem, subscribe_schedule_queue_items


@strawberry.type
class Subscription:
    schedule_auto_mode_state: AsyncIterator[str] = strawberry.field(
        is_subscription=True, resolver=subscribe_auto_mode_state
    )

    schedule_auto_mode_mode: AsyncIterator[str] = strawberry.field(
        is_subscription=True, resolver=subscribe_auto_mode_mode
    )

    schedule_queue_items: AsyncIterator[list[ScheduleQueueItem]] = strawberry.field(
        is_subscription=True, resolver=subscribe_schedule_queue_items
    )



import dataclasses
import itertools
from collections import deque
from collections.abc import Iterable
from datetime import datetime
from typing import Optional


@dataclasses.dataclass
class QueueItem:
    id: int
    name: str
    created_at: datetime
    script: str


@dataclasses.dataclass
class PushArg:
    name: str
    script: str

    def to_queue_item(self, id: int) -> QueueItem:
        return QueueItem(
            id=id,
            name=self.name,
            created_at=datetime.utcnow(),
            script=self.script,
        )


class Queue:
    def __init__(self, items: Optional[Iterable[QueueItem]] = None):
        items = list(items or [])
        id_start = max(i.id for i in items) + 1 if items else 0
        self._id_counter = itertools.count(id_start).__next__
        self.items = deque(items)
        self._map = {i.id: i for i in items}

    def __len__(self) -> int:
        return len(self.items)

    def push(self, arg: PushArg) -> QueueItem:
        item = arg.to_queue_item(id=self._id_counter())
        self.items.append(item)
        self._map[item.id] = item
        return item

    def pop(self) -> QueueItem | None:
        try:
            item = self.items.popleft()
            del self._map[item.id]
            return item
        except IndexError:
            return None

    def remove(self, id: int) -> bool:
        item = self._map.get(id)
        if item is None:
            return False
        self.items.remove(item)
        del self._map[id]
        return True
